import React from 'react';
import { CompositeLayer, Layer } from 'deck.gl';
import { LineLayer, ScatterplotLayer, TextLayer } from '@deck.gl/layers';
import { CompositeLayerType, LayerProps } from '../../mapvis.types';
import { BrushingExtension, CollisionFilterExtension } from '@deck.gl/extensions';

export class NodeLinkLayer extends CompositeLayer<CompositeLayerType> {
  static type = 'nodelink';
  private _layers: Record<string, Layer> = {};

  constructor(props: LayerProps) {
    super(props);
  }

  updateState({ changeFlags }: { changeFlags: any }) {
    return changeFlags.propsOrDataChanged || changeFlags.somethingChanged;
  }

  renderLayers() {
    const { data, settings, getNodeLocation, ml, graphMetadata, selected } = this.props;
    const layerSettings = settings[NodeLinkLayer.type];

    const brushingExtension = new BrushingExtension();
    const collisionFilter = new CollisionFilterExtension();

    const nodeLocations = data.nodes.reduce((acc: Record<string, [number, number]>, node: any) => {
      const pos = getNodeLocation(node._id);
      if (pos && (pos[0] !== 0 || pos[1] !== 0)) {
        acc[node._id] = pos;
      }
      return acc;
    }, {});

    graphMetadata.edges.labels.forEach((label: string) => {
      const layerId = `${label}-edges-line`;

      const edgeData = data.edges.filter((edge: any) => {
        const from = nodeLocations[edge.from];
        const to = nodeLocations[edge.to];
        return from && to;
      });

      this._layers[layerId] = new LineLayer({
        id: layerId,
        data: edgeData,
        visible: !layerSettings?.edges[label]?.hidden,
        pickable: true,
        getWidth: layerSettings?.edges[label]?.width,
        getSourcePosition: (d) => getNodeLocation(d.from),
        getTargetPosition: (d) => getNodeLocation(d.to),
        getColor: (d) => layerSettings?.edges[d.label]?.color,
        extensions: [brushingExtension],
        brushingEnabled: layerSettings?.enableBrushing,
      });
    });

    if (ml?.linkPrediction?.enabled) {
      this._layers['link_prediction'] = new LineLayer({
        id: 'link-prediction-layer',
        data: ml.linkPrediction.result,
        pickable: false,
        getWidth: 1,
        getSourcePosition: (d) => getNodeLocation(d.from),
        getTargetPosition: (d) => getNodeLocation(d.to),
        getColor: (d) => [0, 0, 0],
      });
    }

    graphMetadata.nodes.labels.forEach((label: string) => {
      const layerId = `${label}-nodes-scatterplot`;
      const textLayerId = `${label}-label-target`;

      const nodes = data.nodes.filter((node: any) => nodeLocations[node._id] && node.label === label);

      this._layers[layerId] = new ScatterplotLayer({
        id: layerId,
        visible: !layerSettings?.nodes[label]?.hidden,
        data: nodes,
        pickable: true,
        getFillColor: (d) => layerSettings?.nodes[label]?.color,
        getPosition: (d) => getNodeLocation(d._id),
        getRadius: (d) => layerSettings?.nodes[label]?.size,
        radiusMinPixels: 5,
        getLineWidth: (d: any) => (selected && selected.some((sel) => sel._id === d._id) ? 2 : 1),
        lineWidthUnits: 'pixels',
        stroked: true,
        updateTriggers: {
          getIcon: [selected],
        },
      });

      this._layers[textLayerId] = new TextLayer({
        id: textLayerId,
        data: nodes,
        getPosition: (d: any) => getNodeLocation(d._id),
        getText: (d: any) => d.label,
        getSize: (d: any) => (layerSettings?.nodes[label]?.size * 2) / d.label.length,
        getAlignmentBaseline: 'center',
        getRadius: 10,
        radiusScale: 20,
        getColor: [255, 255, 255],
        sizeUnits: 'meters',
        sizeMaxPixels: 64,
        characterSet: 'auto',
        fontFamily: 'monospace',
        billboard: false,
        getAngle: () => 0,
        collisionGroup: 'textLabels',
        extensions: [collisionFilter],
        collisionEnabled: layerSettings?.collisionEnabled ?? true,
        getCollisionPriority: () => 100,
        collisionTestProps: { sizeScale: 5 },
      });
    });

    return Object.values(this._layers);
  }
}

NodeLinkLayer.layerName = 'NodeLink';
