import React, { useEffect, useCallback, useState, useRef, forwardRef, useImperativeHandle } from 'react';
import DeckGL, { DeckGLProps, DeckGLRef } from '@deck.gl/react';
import { CompositeLayer, FlyToInterpolator, MapViewState, MapViewState, WebMercatorViewport } from '@deck.gl/core';
import { CompositeLayerType, Coordinate, LayerSettingsType, LocationInfo, SearchResultType } from './mapvis.types';
import { VISComponentType, VisualizationPropTypes } from '../../common';
import { layerTypes, createBaseMap, LayerTypes } from './layers';
import { Node } from '@graphpolaris/shared/lib/data-access/store/graphQueryResultSlice';
import { geoCentroid } from 'd3';
import { Attribution, ActionBar, MapTooltip, MapSettings } from './components';
import { Tooltip, TooltipContent, TooltipTrigger } from '@graphpolaris/shared/lib/components';
import { useSelectionLayer, useCoordinateLookup } from './hooks';

export type MapProps = {
  layer: LayerTypes;
  location: Record<string, LocationInfo>;
} & Partial<Record<LayerTypes, LayerSettingsType>>;

const settings: MapProps = {
  layer: 'nodelink',
  location: {},
};

const INITIAL_VIEW_STATE: MapViewState = {
  latitude: 52.1006,
  longitude: 5.6464,
  zoom: 6,
  bearing: 0,
  pitch: 0,
};

const FLY_SPEED = 1000;

export const MapVis = forwardRef((props: VisualizationPropTypes<MapProps>, refExternal) => {
  const ref = useRef<HTMLDivElement>(null);
  const baseLayer = useRef(createBaseMap());
  const [viewport, setViewport] = useState<Record<string, any>>(INITIAL_VIEW_STATE);
  const [hoverObject, setHoverObject] = useState<Node | null>(null);
  const [selected, setSelected] = useState<any[]>([]);
  const [selectingRectangle, setSelectingRectangle] = useState<boolean>(false);
  const [layerIds, setLayerIds] = useState<string[]>([]);
  const [isSearching, setIsSearching] = useState<boolean>(false);
  const [searchResult, setSearchResult] = useState<SearchResultType | undefined>(undefined);

  const coordinateLookup = useCoordinateLookup(props.data.nodes, props.settings.location);
  const [shouldExport, setShouldExport] = useState<boolean>(false);

  const captureImage = () => {
    const canvas = ref.current?.querySelector('canvas');

    if (canvas) {
      canvas.toBlob((blob) => {
        if (blob) {
          const link = document.createElement('a');
          link.download = 'map-visualization.png';
          link.href = URL.createObjectURL(blob);
          link.click();
        } else {
          console.error('Failed to capture canvas');
        }
      }, 'image/png');
    } else {
      console.error('Canvas element not found');
    }
  };

  const exportImageInternal = () => {
    requestAnimationFrame(() => {
      captureImage();
      props.updateSettings({ export2PNG: false });
    });
  };
  // !FIXME I have to wrap exportImageInternal() with a useEffect otherwise didnt work, by just calling exportImageInternal() from outside
  useEffect(() => {
    if (props.settings.export2PNG) {
      exportImageInternal();
    }
  }, [props.settings.export2PNG]);

  useEffect(() => {
    if (shouldExport) {
      exportImageInternal();
      setShouldExport(false);
    }
  }, [shouldExport]);

  useImperativeHandle(refExternal, () => ({
    exportImageInternal: () => setShouldExport(true),
  }));

  const getFittedViewport = useCallback(
    (minLat: number, maxLat: number, minLon: number, maxLon: number) => {
      const viewportWebMercator = new WebMercatorViewport(viewport).fitBounds(
        [
          [minLon, minLat],
          [maxLon, maxLat],
        ],
        { padding: 20 },
      );
      const { zoom, longitude, latitude } = viewportWebMercator;
      return { zoom, longitude, latitude };
    },
    [viewport],
  );

  const flyToBoundingBox = useCallback(
    (minLat: number, maxLat: number, minLon: number, maxLon: number, options = {}) => {
      const fittedViewport = getFittedViewport(minLat, maxLat, minLon, maxLon);
      setViewport((prevViewport) => ({
        ...prevViewport,
        ...options,
        ...fittedViewport,
        transitionDuration: FLY_SPEED,
        transitionInterpolator: new FlyToInterpolator(),
      }));
    },
    [getFittedViewport],
  );

  const [dataLayer, setDataLayer] = useState<{ component: CompositeLayer<CompositeLayerType>; id: string } | null>(null);

  useEffect(() => {
    if (!props.settings.layer) {
      setDataLayer(null);
      return;
    }

    const layer = {
      id: props.settings.layer,
      component: props.settings.layer ? layerTypes?.[props.settings.layer] : layerTypes.nodelink,
    };

    const layerProps: CompositeLayerType = {
      ...props,
      selected: selected,
      hoverObject: hoverObject,
      getNodeLocation: (d: string) => coordinateLookup[d],
      flyToBoundingBox: flyToBoundingBox,
      setLayerIds: (val: string[]) => setLayerIds(val),
    };

    if (dataLayer && dataLayer.id === layer.id) {
      setDataLayer({ component: dataLayer.component.clone(layerProps), id: props.settings.layer });
    } else {
      // @ts-ignore
      setDataLayer({ component: new layer.component(layerProps), id: props.settings.layer });
    }
  }, [props.settings.layer, props.data, selected, hoverObject, props.settings]);

  const selectionLayer = useSelectionLayer(selectingRectangle, layerIds, (pickingInfos: any[]) => {
    const nodes: Node[] = [];
    const edges: any[] = [];
    pickingInfos.forEach(({ object }) => {
      if (object._id) {
        if (object.from && object.to) {
          edges.push(object);
        } else {
          nodes.push(object);
        }
      }
    });
    setSelected(nodes.map((node) => node._id));
    props.handleSelect({ nodes, edges });
    setSelectingRectangle(false);
  });

  const coordinateToXY = useCallback(
    (coordinate: Coordinate) => {
      const [longitude, latitude] = coordinate;
      return new WebMercatorViewport(viewport).project([longitude, latitude]);
    },
    [viewport],
  );

  useEffect(() => {
    if (selected.length > 0) {
      const updatedSelected = selected.map((node) => {
        let x, y;
        if (props.settings.layer === 'nodelink' && node.lon && node.lat) {
          const coordinate: Coordinate = [parseFloat(node.lon), parseFloat(node.lat)];
          [x, y] = coordinateToXY(coordinate);
        } else if (props.settings.layer === 'choropleth') {
          const centroid = geoCentroid(node);
          if (centroid) {
            [x, y] = coordinateToXY([centroid[0], centroid[1]]);
          }
        }
        return { ...node, x, y };
      });
      setSelected(updatedSelected);
    }
  }, [viewport]);

  const handleClick = useCallback(
    (object: any) => {
      if (props.data) {
        if (!object) {
          props.handleSelect();
          setSelected([]);
          setSearchResult(undefined);
          return;
        }
        if (object.hasOwnProperty('attributes') && object.hasOwnProperty('id') && object.hasOwnProperty('label')) {
          const objectLocation: Coordinate = coordinateLookup[object.id];
          props.handleSelect({ nodes: [object] });
          if (objectLocation) {
            const [x, y] = coordinateToXY(objectLocation);
            setSelected([{ ...object, x, y, lon: objectLocation[0], lat: objectLocation[1], selectedType: 'node' }]);
          }
        }
        if (object.type === 'Feature') {
          const centroid = geoCentroid(object);
          if (centroid) {
            const [x, y] = coordinateToXY(centroid);
            setSelected([{ ...object, x, y, selectedType: 'area' }]);
          }
          const ids = object.properties.nodes;
          if (ids && ids.length > 0) {
            const nodes = props.data.nodes.filter((node) => ids.includes((node as unknown as { id: string }).id));
            props.handleSelect({ nodes: [...nodes] });
          }
        }
      }
    },
    [props, coordinateLookup, coordinateToXY],
  );

  useEffect(() => {
    if (searchResult) {
      const coordinate: Coordinate = [parseFloat(searchResult.lon), parseFloat(searchResult.lat)];
      const [x, y] = coordinateToXY(coordinate);

      if (searchResult.x !== x || searchResult.y !== y) {
        setSearchResult((prev) => (prev ? { ...prev, x, y } : undefined));
      }
    }
  }, [viewport, searchResult, coordinateToXY]);

  return (
    <div className="w-full h-full flex-grow relative overflow-hidden" ref={ref}>
      <DeckGL
        layers={[baseLayer.current, dataLayer?.component, selectionLayer]}
        controller={true}
        initialViewState={viewport}
        onViewStateChange={({ viewState }) => setViewport(viewState)}
        onClick={({ object }) => handleClick(object)}
        onHover={({ object }) => setHoverObject(object !== undefined ? object : null)}
      />
      <ActionBar
        isSearching={isSearching}
        setIsSearching={setIsSearching}
        setSearchResult={setSearchResult}
        setSelectingRectangle={setSelectingRectangle}
        flyToBoundingBox={flyToBoundingBox}
      />
      {selected.length > 0 &&
        selected.map((node, index) => (
          <Tooltip key={index} open={true} interactive={false} boundaryElement={ref} showArrow={true}>
            <TooltipTrigger x={node.x} y={node.y} />
            <TooltipContent>
              <MapTooltip type={node.selectedType} onClose={() => {}} data={{ ...node }} key={node._id} />
            </TooltipContent>
          </Tooltip>
        ))}
      {searchResult && (
        <Tooltip open={true} interactive={false} boundaryElement={ref} showArrow={true}>
          <TooltipTrigger x={searchResult.x} y={searchResult.y} />
          <TooltipContent>
            <MapTooltip type="location" onClose={() => {}} data={{ ...searchResult }} key={searchResult.name} />
          </TooltipContent>
        </Tooltip>
      )}
      <Attribution />
    </div>
  );
});

const mapRef = React.createRef<{ exportImageInternal: () => void }>();

const MapComponent: VISComponentType<MapProps> = {
  displayName: 'MapVis',
  description: 'Geographical Features',
  component: React.forwardRef((props: VisualizationPropTypes<MapProps>, ref) => <MapVis {...props} ref={mapRef} />),
  settingsComponent: MapSettings,
  settings: settings,
  exportImage: () => {
    if (mapRef.current) {
      mapRef.current.exportImageInternal();
    } else {
      console.error('Map reference is not set.');
    }
  },
};
export default MapComponent;
